﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using ThinkGeo.MapSuite.Core;

namespace SplitPolygon
{
    public static class Splitter
    {
        public static Collection<LineShape> GetPerpendicularLines(LineShape lineShape, Collection<PointShape> interceptedPoints, double length)
        {
            Collection<LineShape> returnFeatures = new Collection<LineShape>();
            double halfLength = length / 2;
            double pi2 = Math.PI / 2;
            double searchingTolerance = 0.000001;
            foreach (PointShape interceptedPoint in interceptedPoints)
            {
                RectangleShape searchingArea = new RectangleShape(interceptedPoint.X - searchingTolerance, interceptedPoint.Y + searchingTolerance, interceptedPoint.X + searchingTolerance, interceptedPoint.Y - searchingTolerance);

                // Check if the intercepted point is the vertex will be ignore, because we can not create perpendicular line crosses vertex.
                foreach (Vertex vertex in lineShape.Vertices)
                {
                    if (new PointShape(vertex).IsWithin(searchingArea))
                    {
                        continue;
                    }
                }

                double middleX = interceptedPoint.X;
                double middleY = interceptedPoint.Y;

                Vertex vertex1 = new Vertex();
                Vertex vertex2 = new Vertex();
                // Find the crossing point is between which two vertices
                for (int i = 0; i < lineShape.Vertices.Count - 1; i++)
                {
                    vertex1 = lineShape.Vertices[i];
                    vertex2 = lineShape.Vertices[i + 1];
                    LineShape currentLine = new LineShape(new Vertex[] { vertex1, vertex2 });
                    if (searchingArea.Intersects(currentLine))
                    {
                        break;
                    }
                }

                double angle = Math.Atan((vertex1.Y - vertex2.Y) / (vertex1.X - vertex2.X));

                double x = 0;
                double y = 0;

                if (angle > 0 && angle < pi2)
                {
                    if (vertex1.X > vertex2.X)
                    {
                        x = middleX + Math.Sin(angle) * halfLength;
                        y = middleY - Math.Cos(angle) * halfLength;
                    }
                    else
                    {
                        x = middleX - Math.Sin(angle) * halfLength;
                        y = middleY + Math.Cos(angle) * halfLength;
                    }
                }
                else
                {
                    if (vertex1.X > vertex2.X)
                    {
                        x = middleX + Math.Sin(angle) * halfLength;
                        y = middleY - Math.Cos(angle) * halfLength;
                    }
                    else
                    {
                        x = middleX - Math.Sin(angle) * halfLength;
                        y = middleY + Math.Cos(angle) * halfLength;
                    }
                }

                Vertex anotherVertex = new Vertex(middleX * 2 - x, middleY * 2 - y);
                LineShape line = new LineShape(new Vertex[] { anotherVertex, new Vertex(x, y) });
                returnFeatures.Add(line);
            }

            return returnFeatures;
        }

        public static PolygonShape SplitPolygonWithTwoLines(PolygonShape processedPolygon, LineShape lineShape1, LineShape lineShape2)
        {
            MultipolygonShape resultMultipolygon = Split(processedPolygon, lineShape1);
            PolygonShape resultShape = new PolygonShape();
            foreach (PolygonShape polygonShape in resultMultipolygon.Polygons)
            {
                MultipointShape intersectionMultiPoint = lineShape2.GetCrossing(polygonShape.OuterRing);
                if (intersectionMultiPoint.Points.Count == 2)
                {
                    MultipolygonShape resultMultipolygon1 = Split(polygonShape, lineShape2);
                    foreach (PolygonShape polygonShape1 in resultMultipolygon1.Polygons)
                    {
                        MultipointShape intersectionMultiPoint1 = lineShape2.GetCrossing(polygonShape.OuterRing);
                        if (intersectionMultiPoint1.Points.Count == 2)
                        {
                            resultShape = polygonShape1;
                            break;
                        }
                    }
                }
            }
            return resultShape;
        }

        public static MultipolygonShape Split(PolygonShape processedPolygon, LineShape lineShape)
        {
            MultipolygonShape resultShape = new MultipolygonShape();
            MultipointShape intersectionMultiPoint = lineShape.GetCrossing(processedPolygon.OuterRing);

            if (intersectionMultiPoint.Points.Count == 2)
            {
                PolygonShape polygonShape1 = GetPolygonForSplit(processedPolygon, lineShape, false);
                PolygonShape polygonShape2 = GetPolygonForSplit(processedPolygon, lineShape, true);
                resultShape.Polygons.Add(polygonShape1);
                resultShape.Polygons.Add(polygonShape2);
            }
            return resultShape;
        }

        private static PolygonShape GetPolygonForSplit(PolygonShape processedPolygon, LineShape processedLineShape, bool changeOrder)
        {
            MultipointShape intersectionMultiPoint = new MultipointShape();
            PolygonShape orderedPolygon = processedPolygon;

            LineShape lineShape = new LineShape((processedLineShape.GetIntersection(orderedPolygon.OuterRing)).Lines[0].GetWellKnownText());
            if (changeOrder)
            {
                List<Vertex> verticesToReverse = new List<Vertex>(lineShape.Vertices);
                verticesToReverse.Reverse();
                lineShape.Vertices.Clear();
                foreach (Vertex vertice in verticesToReverse)
                { lineShape.Vertices.Add(vertice); }
            }

            foreach (Vertex vertex in lineShape.Vertices)
            { intersectionMultiPoint.Points.Add(new PointShape(vertex)); }

            if (intersectionMultiPoint.Points.Count >= 2)
            {
                PolygonShape resultPolygonShape = new PolygonShape();
                RingShape resultOuterRing = SplitRing(orderedPolygon.OuterRing, intersectionMultiPoint.Points);
                resultPolygonShape.OuterRing = resultOuterRing;

                foreach (RingShape innerRing in orderedPolygon.InnerRings)
                {
                    MultipointShape innerIntersectionMultiPoint = lineShape.GetCrossing(innerRing);
                    if (innerIntersectionMultiPoint.Points.Count == 2)
                    {
                        RingShape resultInnerRing = SplitRing(innerRing, innerIntersectionMultiPoint.Points);
                        if (resultPolygonShape.Contains(resultInnerRing))
                        {
                            resultPolygonShape.InnerRings.Add(resultInnerRing);
                        }
                    }
                    else
                    {
                        if (resultPolygonShape.Contains(innerRing))
                        {
                            resultPolygonShape.InnerRings.Add(innerRing);
                        }
                    }
                }
                return resultPolygonShape;
            }
            return null;
        }

        public static bool DoesPointShapeBelongToLineSegment(PointShape pointShape, PointShape linePointShape1, PointShape linePointShape2)
        {
            bool result = false;
            double a = 0;
            double b = 0;

            if ((pointShape.X == linePointShape1.X & pointShape.Y == linePointShape1.Y) | (pointShape.X == linePointShape2.X & pointShape.Y == linePointShape2.Y))
            {
                result = true;
            }
            else
            {
                if (linePointShape1.X != linePointShape2.X)
                {
                    a = (linePointShape2.Y - linePointShape1.Y) / (linePointShape2.X - linePointShape1.X);
                    b = linePointShape1.Y - (a * linePointShape1.X);

                    if (Math.Round(pointShape.Y, 5) == Math.Round((a * pointShape.X) + b, 5) & pointShape.X >= Math.Min(linePointShape1.X, linePointShape2.X) & pointShape.X <= Math.Max(linePointShape1.X, linePointShape2.X))
                    {
                        result = true;
                    }
                    else
                    {
                        result = false;
                    }
                }
                else
                {
                    if (pointShape.X == linePointShape1.X & (pointShape.Y >= Math.Min(linePointShape1.Y, linePointShape2.Y) & pointShape.Y <= Math.Max(linePointShape1.Y, linePointShape2.Y)))
                    {
                        result = true;
                    }
                    else
                    { result = false; }
                }
            }
            return result;
        }

        public static RingShape SplitRing(RingShape processedRing, Collection<PointShape> intersectionPointShapes)
        {
            RingShape resultRingShape = new RingShape();
            PointShape intersectionPointShape1 = intersectionPointShapes[0];
            PointShape intersectionPointShape2 = intersectionPointShapes[intersectionPointShapes.Count - 1];

            int i = 0;
            int totalPointNumber = processedRing.Vertices.Count;
            while (i < totalPointNumber - 1)
            {
                int indexA = i + 1;
                if (DoesPointShapeBelongToLineSegment(intersectionPointShape1, new PointShape(processedRing.Vertices[i]), new PointShape(processedRing.Vertices[indexA])))
                {
                    resultRingShape.Vertices.Add(new Vertex(intersectionPointShape1));
                    if (DoesPointShapeBelongToLineSegment(intersectionPointShape2, new PointShape(processedRing.Vertices[i]), new PointShape(processedRing.Vertices[indexA])))
                    {

                        for (int shapeIndex = intersectionPointShapes.Count - 1; shapeIndex > 0; shapeIndex--)
                        { resultRingShape.Vertices.Add(new Vertex(intersectionPointShapes[shapeIndex])); }
                        resultRingShape.Vertices.Add(new Vertex(intersectionPointShape1));
                    }
                    else
                    {
                        for (int j = i + 1; j <= processedRing.Vertices.Count - 1; j++)
                        {
                            //- 1
                            int indexB = j + 1;

                            if (j < processedRing.Vertices.Count - 1)
                            {
                                if (DoesPointShapeBelongToLineSegment(intersectionPointShape2, new PointShape(processedRing.Vertices[j]), new PointShape(processedRing.Vertices[indexB])))
                                {
                                    resultRingShape.Vertices.Add(processedRing.Vertices[j]);

                                    for (int shapeIndex = intersectionPointShapes.Count - 1; shapeIndex > 0; shapeIndex--)
                                    { resultRingShape.Vertices.Add(new Vertex(intersectionPointShapes[shapeIndex])); }
                                    resultRingShape.Vertices.Add(new Vertex(intersectionPointShape1));
                                    break; // TODO: might not be correct. Was : Exit For
                                }
                                else
                                { resultRingShape.Vertices.Add(processedRing.Vertices[j]); }
                            }
                            else
                            {
                                for (int k = 0; k < i; k++)
                                {
                                    if (DoesPointShapeBelongToLineSegment(intersectionPointShape2, new PointShape(processedRing.Vertices[k]), new PointShape(processedRing.Vertices[(k + 1)])))
                                    {
                                        resultRingShape.Vertices.Add(processedRing.Vertices[k]);
                                        for (int shapeIndex = intersectionPointShapes.Count - 1; shapeIndex > 0; shapeIndex--)
                                        { resultRingShape.Vertices.Add(new Vertex(intersectionPointShapes[shapeIndex])); }
                                        resultRingShape.Vertices.Add(new Vertex(intersectionPointShape1));
                                        break; // TODO: might not be correct. Was : Exit For
                                    }
                                    else
                                    {
                                        resultRingShape.Vertices.Add(processedRing.Vertices[k]);
                                    }
                                }
                                break; // TODO: might not be correct. Was : Exit For
                            }
                        }
                    }

                    return resultRingShape;
                }
                i = i + 1;
            }
            return null;
        }
    }
}
